from typing import Dict, List
from xml.etree import ElementTree

from pycubexr.classes import Metric, Region, CNode, SystemTreeNode
from pycubexr.classes.metric import MetricType
from pycubexr.parsers import xml_parser_helper


class AnchorXMLParseResult(object):

    def __init__(
            self,
            *,
            attrs: Dict[str, str],
            metrics: List[Metric],
            regions: List[Region],
            cnodes: List[CNode],
            system_tree_nodes: List[SystemTreeNode]
    ):
        self.attrs = attrs
        self.metrics = metrics
        self.regions = regions
        self.cnodes = cnodes
        self.system_tree_nodes = system_tree_nodes
        self.regions_by_id = {r.id: r for r in regions}
        self.all_cnodes = {cnode.id: cnode for root_cnode in cnodes for cnode in root_cnode.get_all_children()}


def parse_anchor_xml(root: ElementTree):
    result = AnchorXMLParseResult(
        attrs=xml_parser_helper.parse_attrs(root),
        metrics=xml_parser_helper.parse_metrics(root),
        regions=xml_parser_helper.parse_regions(root),
        cnodes=xml_parser_helper.parse_cnodes(root),
        system_tree_nodes=xml_parser_helper.parse_system_tree_nodes(root),
    )
    assert len(result.cnodes) == 1
    assert len(result.system_tree_nodes) == 1

    _metric_tree_enumerations(result)
    for cnode in result.all_cnodes.values():
        cnode.region = result.regions_by_id[cnode.callee_region_id]
    return result


def _wide_enumeration(root_list: List[CNode]):
    # does not implement breadth first search
    visited = {}
    stack = []
    ctr = 0

    for root_node in root_list:
        visited[ctr] = root_node.id
        ctr += 1
        stack.append(root_node)
        while stack:
            node = stack.pop()
            for child in node.get_children():
                visited[ctr] = child.id
                ctr += 1
            for child in reversed(node.get_children()):
                stack.append(child)
    return visited


def _deep_enumeration(root_list: List[CNode]):
    visited = {}
    stack = []
    ctr = 0
    for root_node in reversed(root_list):
        stack.append(root_node)

    while stack:
        node = stack.pop()
        visited[ctr] = node.id
        ctr += 1
        for child in reversed(node.get_children()):
            stack.append(child)

    return visited


def _metric_tree_enumerations(result):
    deep_enumeration = _deep_enumeration(result.cnodes)
    wide_enumeration = _wide_enumeration(result.cnodes)

    def walk_tree(childs):
        for metric in childs:
            if metric.metric_type == MetricType.EXCLUSIVE:
                metric.tree_index_to_cid_map = deep_enumeration
            elif metric.metric_type == MetricType.INCLUSIVE:
                metric.tree_index_to_cid_map = wide_enumeration
            walk_tree(metric.childs)

    walk_tree(result.metrics)


def _assign_region(cnode, result):
    cnode.region = [r for r in result.regions if r.id == cnode.callee_region_id][0]
    for c in cnode.get_children():
        _assign_region(c, result)



import struct
import tarfile
import warnings

BLOCKSIZE = 512
NUL = b"\0"


def nti(s):
    """Convert a number field to a python number.
    """
    # Copyright (C) 2002 Lars Gustaebel <lars@gustaebel.de>
    # All rights reserved.
    #
    # Permission  is  hereby granted,  free  of charge,  to  any person
    # obtaining a  copy of  this software  and associated documentation
    # files  (the  "Software"),  to   deal  in  the  Software   without
    # restriction,  including  without limitation  the  rights to  use,
    # copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies  of  the  Software,  and to  permit  persons  to  whom the
    # Software  is  furnished  to  do  so,  subject  to  the  following
    # conditions:
    #
    # The above copyright  notice and this  permission notice shall  be
    # included in all copies or substantial portions of the Software.
    #
    # THE SOFTWARE IS PROVIDED "AS  IS", WITHOUT WARRANTY OF ANY  KIND,
    # EXPRESS OR IMPLIED, INCLUDING  BUT NOT LIMITED TO  THE WARRANTIES
    # OF  MERCHANTABILITY,  FITNESS   FOR  A  PARTICULAR   PURPOSE  AND
    # NONINFRINGEMENT.  IN  NO  EVENT SHALL  THE  AUTHORS  OR COPYRIGHT
    # HOLDERS  BE LIABLE  FOR ANY  CLAIM, DAMAGES  OR OTHER  LIABILITY,
    # WHETHER  IN AN  ACTION OF  CONTRACT, TORT  OR OTHERWISE,  ARISING
    # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    # OTHER DEALINGS IN THE SOFTWARE.

    # There are two possible encodings for a number field, see
    # itn() below.
    if s[0] in (0o200, 0o377):
        n = 0
        for i in range(len(s) - 1):
            n <<= 8
            n += s[i + 1]
        if s[0] == 0o377:
            n = -(256 ** (len(s) - 1) - n)
    else:
        try:
            p = s.find(b"\0")
            if p != -1:
                s = s[:p]
            s = s.decode("ascii", "strict")
            n = int(s.strip() or "0", 8)
        except ValueError:
            raise ValueError("invalid tar header")
    return n


class TarInfoWithoutCheck(tarfile.TarInfo):
    checksum_warning = "Detected invalid checksum in CUBE file header. Proceeding without checking checksums. " \
                       "If CubeWriter 4.8 was used to create this file, you can ignore this warning."

    @classmethod
    def frombuf(cls, buf, encoding, errors):
        if len(buf) == 0:
            super(TarInfoWithoutCheck, cls).frombuf(buf, encoding, errors)
        if len(buf) != BLOCKSIZE:
            super(TarInfoWithoutCheck, cls).frombuf(buf, encoding, errors)
        if buf.count(tarfile.NUL) == BLOCKSIZE:
            super(TarInfoWithoutCheck, cls).frombuf(buf, encoding, errors)

        chksum = nti(buf[148:156])
        unsigned_chksum = 256 + sum(struct.unpack_from("148B8x356B", buf))
        signed_chksum = 256 + sum(struct.unpack_from("148b8x356b", buf))
        calculated_checksums = unsigned_chksum, signed_chksum
        if chksum not in calculated_checksums:
            warnings.warn(cls.checksum_warning)
        buf = buf[:148] + bytes("%07o\0" % unsigned_chksum, "ascii") + buf[156:]

        return super(TarInfoWithoutCheck, cls).frombuf(buf, encoding, errors)



from pycubexr.classes import Metric


class MissingMetricError(Exception):
    def __init__(self, metric: Metric):
        self.metric = metric
        super().__init__('The cubex file does NOT contain an index and/or data for the metric ({})'.format(metric))


class CorruptIndexError(Exception):
    def __init__(self, message='Corrupt index.'):
        super().__init__(message)


class InvalidConversionInstructionError(Exception):
    def __init__(self, message="Received invalid conversion instruction. "
                               "convert_to_exclusive and convert_to_inclusive must not be True at the same time."):
        super().__init__(message)



from typing import BinaryIO

from pycubexr.classes import MetricValues, Metric
from pycubexr.parsers.data_parser import parse_data
from pycubexr.parsers.index_parser import parse_index


def extract_metric_values(
        *,
        metric: Metric,
        index_file: BinaryIO,
        data_file: BinaryIO,
        allow_full_uint64_values: bool = False
) -> MetricValues:
    index = parse_index(index_file=index_file)
    values = parse_data(
        data_file=data_file,
        data_type=metric.data_type,
        endianness_format_char=index.endianness_format,
        allow_full_uint64_values=allow_full_uint64_values,
    )
    cnode_indices = [metric.tree_index_to_cid_map[tidx] for tidx in index.tree_indices]
    return MetricValues(
        metric=metric,
        cnode_indices=cnode_indices,
        values=values
    )



import numpy

METRIC_FORMATS = {
    'CHAR': numpy.uint8,
    'COMPLEX': numpy.complex128,
    'DOUBLE': numpy.float64,
    'FLOAT': numpy.float64,  # Is defined as float 64
    # 'HISTOGRAM': lambda n: str(int(n)) + 'd',
    'INT': numpy.int32,
    'INT16': numpy.int16,
    'INT32': numpy.int32,
    'INT64': numpy.int64,
    'INT8': numpy.int8,
    'INTEGER': numpy.int64,
    'MAXDOUBLE': numpy.float64,
    'MINDOUBLE': numpy.float64,
    'NDOUBLES': lambda n: (numpy.float64, (n,)),
    'RATE': [('main', numpy.float64), ('time', numpy.float64)],
    # 'SCALE_FUNC': 'P',
    'SHORT INT': numpy.int16,
    'SIGNED INT': numpy.int32,
    'SIGNED INTEGER': numpy.int64,
    'SIGNED SHORT INT': numpy.int16,
    'TAU_ATOMIC': [('n', numpy.uint32), ('min', numpy.float64), ('max', numpy.float64), ('sum', numpy.float64),
                   ('sum2', numpy.float64)],  # TODO check if uint32 is correct
    'UINT16': numpy.uint16,
    'UINT32': numpy.uint32,
    'UINT64': numpy.uint64,
    'UINT8': numpy.uint8,
    'UNSIGNED INT': numpy.uint32,
    'UNSIGNED INTEGER': numpy.uint64,
    'UNSIGNED SHORT INT': numpy.uint16,
}



import unittest
from pathlib import Path

from pycubexr import CubexParser


class TestHWCounterMeasurements1(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cubex_file_path = Path("../data/hw_counter/L2_DCM_negative.cubex").resolve()
        cls.cubex = CubexParser(cubex_file_path).__enter__()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cubex.__exit__(None, None, None)

    def get_values_for_metric(self, name, convert_to_exclusive=False, convert_to_inclusive=False):
        metric = self.cubex.get_metric_by_name(name)
        metric_values = self.cubex.get_metric_values(metric)
        # iterate over all callpaths in cubex file
        cnode_values = []
        for cnode in self.cubex.all_cnodes():
            # return the measurement values for all mpi processes for the current metric and callpath
            value = metric_values.value(cnode, convert_to_exclusive, convert_to_inclusive)
            cnode_values.append(value)
        return cnode_values

    def test_PAPI_L2_DCM_exclusive(self):
        metric_values = self.get_values_for_metric('PAPI_L2_DCM', convert_to_exclusive=True)
        for cnode_value in metric_values:
            self.assertGreaterEqual(cnode_value, 0)
            self.assertLess(cnode_value, 0xFFFF_FFFF_FFFF_FFFF)

    def test_PAPI_L2_DCM_inclusive(self):
        metric_values = self.get_values_for_metric('PAPI_L2_DCM', convert_to_inclusive=True)
        for cnode_value in metric_values:
            self.assertGreaterEqual(cnode_value, 0)
            self.assertLess(cnode_value, 0xFFFFFFFFFFFFFFFF)

    def test_PAPI_L2_DCM_allow_full_uint64(self):
        metric = self.cubex.get_metric_by_name('PAPI_L2_DCM')
        metric_values = self.cubex.get_metric_values(metric,allow_full_uint64_values=True)
        self.assertTrue(any(metric_values.values >= 0xFFFF_FFFF_FFFF_FFFF))
        check = False
        for cnode in self.cubex.all_cnodes():
            check = check or any(metric_values.cnode_values(cnode) >= 0xFFFF_FFFF_FFFF_FFFF)
        self.assertTrue(check)
        values = [metric_values.value(cnode) for cnode in self.cubex.all_cnodes()]
        self.assertTrue(all(v >= 0 for v in values))
        self.assertTrue(any(v >= 0xFFFF_FFFF_FFFF_FFFF for v in values))


class TestHWCounterMeasurements2(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cubex_file_path = Path("../data/hw_counter/L3_DCA_negative.cubex").resolve()
        cls.cubex = CubexParser(cubex_file_path).__enter__()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cubex.__exit__(None, None, None)

    def get_values_for_metric(self, name, convert_to_exclusive=False, convert_to_inclusive=False):
        metric = self.cubex.get_metric_by_name(name)
        metric_values = self.cubex.get_metric_values(metric)
        # iterate over all callpaths in cubex file
        cnode_values = []
        for cnode in self.cubex.all_cnodes():
            # return the measurement values for all mpi processes for the current metric and callpath
            value = metric_values.value(cnode, convert_to_exclusive, convert_to_inclusive)
            self.assertGreaterEqual(value, 0)
            cnode_values.append(value)
        return cnode_values

    def test_PAPI_L2_DCM_exclusive(self):
        metric_values = self.get_values_for_metric('PAPI_L3_DCA', convert_to_exclusive=True)
        for cnode_value in metric_values:
            self.assertGreaterEqual(cnode_value, 0)
            self.assertLess(cnode_value, 0xFFFFFFFFFFFFFFFF)

    def test_PAPI_L2_DCM_inclusive(self):
        metric_values = self.get_values_for_metric('PAPI_L3_DCA', convert_to_inclusive=True)
        for cnode_value in metric_values:
            self.assertGreaterEqual(cnode_value, 0)
            self.assertLess(cnode_value, 0xFFFFFFFFFFFFFFFF)


if __name__ == '__main__':
    unittest.main()



import logging
import warnings
from typing import List

import numpy as np

from pycubexr.classes import CNode, Metric
from pycubexr.classes.metric import MetricType
from pycubexr.classes.values import CubeValues
from pycubexr.utils.exceptions import InvalidConversionInstructionError


class MetricValues(object):

    def __init__(
            self,
            *,
            metric: Metric,
            cnode_indices: List[int],
            values: np.ndarray
    ):
        self.metric = metric
        self.values = values
        self.cnode_indices = {cid: i for i, cid in enumerate(cnode_indices)}
        self._num_locations = int(len(self.values) / len(self.cnode_indices))
        assert len(self.values) % len(self.cnode_indices) == 0

    def num_locations(self):
        return self._num_locations

    def cnode_values(self, cnode: CNode, convert_to_exclusive: bool = False, convert_to_inclusive: bool = False):
        if isinstance(cnode, int):
            warnings.warn('Calling with Cnode ID is deprecated use Cnode directly.', DeprecationWarning)
            if convert_to_inclusive or convert_to_exclusive:
                raise InvalidConversionInstructionError('Conversion is not supported when passing a Cnode ID.')
            cnode = CNode(_id=cnode, callee_region_id=-1)
        else:
            if convert_to_inclusive and convert_to_exclusive:
                raise InvalidConversionInstructionError()
            assert not (convert_to_inclusive and convert_to_exclusive)

        cid = cnode.id
        if cid not in self.cnode_indices:
            values = np.zeros(self.num_locations(), dtype=self.values.dtype)
        else:
            start_index = int(self.cnode_indices[cid] * self.num_locations())
            end_index = start_index + self.num_locations()
            values = self.values[start_index: end_index]  # creates no copy

        must_convert = ((convert_to_exclusive and self.metric.metric_type == MetricType.INCLUSIVE)
                        or (convert_to_inclusive and self.metric.metric_type == MetricType.EXCLUSIVE))
        if must_convert:
            values = self._convert_values(cnode, values, to_inclusive=convert_to_inclusive)
        return values

    def location_value(self, cnode: CNode, location_id: int, convert_to_inclusive=False, convert_to_exclusive=False):
        assert location_id < self.num_locations()
        return self.cnode_values(
            cnode,
            convert_to_exclusive=convert_to_exclusive,
            convert_to_inclusive=convert_to_inclusive
        )[location_id]

    @staticmethod
    def _detect_negative_overflow(a, b):
        if np.issubdtype(a.dtype, np.unsignedinteger):
            check = a < b
            if np.any(check):
                b = b.copy()
                b[check] = a[check]
        elif np.issubdtype(a.dtype, np.signedinteger):
            pre_check = a > 0
            check = np.full_like(pre_check, False, dtype=bool)
            check[~pre_check] = -b[~pre_check] < np.iinfo(a.dtype).min - a[~pre_check]
            check[pre_check] |= -b[pre_check] > np.iinfo(a.dtype).max - a[pre_check]
            if np.any(check):
                if not np.issubdtype(a.dtype, np.int64):
                    bigger_type = np.min_scalar_type(int(np.iinfo(a.dtype).min) - 1)
                    b = b.astype(bigger_type, copy=False)
                else:
                    b = b.copy()
                    b[~pre_check & check] = -(np.iinfo(a.dtype).min - a[~pre_check & check])
                    b[pre_check & check] = -(np.iinfo(a.dtype).max - a[pre_check & check])
        return b

    def _convert_values(self, cnode: CNode, values: np.ndarray, to_inclusive: bool = True):
        # Go over all cnode children and add the metric values
        values = values.copy()
        for child_cnode in cnode.get_children():
            child_values = self.cnode_values(child_cnode,
                                             convert_to_inclusive=True,
                                             convert_to_exclusive=False)
            if to_inclusive:
                values += child_values
            else:
                if isinstance(values, np.ndarray) and np.issubdtype(values.dtype, np.integer):
                    child_values = self._detect_negative_overflow(values, child_values)
                values -= child_values
        return values

    def value(self, cnode: CNode, convert_to_exclusive: bool = False, convert_to_inclusive: bool = False):
        res = self.cnode_values(cnode, convert_to_exclusive, convert_to_inclusive)
        sum_ = res.sum()
        if isinstance(res, np.ndarray) and np.issubdtype(res.dtype, np.integer):
            if res.max() > np.iinfo(res.dtype).max // len(res):
                logging.info("Used overflow prevention for {0} of {1}".format(self.metric.name, cnode.region.name))
                # Overflow prevention, really slow
                sum_ = 0
                for r in res:
                    sum_ += int(r)

        if isinstance(sum_, CubeValues):
            return sum_.astype(float)
        else:
            return sum_

    def mean(self, cnode: CNode, convert_to_exclusive: bool = False, convert_to_inclusive: bool = False):
        res = self.cnode_values(cnode, convert_to_exclusive, convert_to_inclusive).mean()
        if isinstance(res, CubeValues):
            return res.astype(float)
        else:
            return res

    def __repr__(self):
        return 'MetricValues<{}>'.format(self.__dict__)



import setuptools

with open("README.md") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pycubexr",
    version="2.0.1",
    author="Extra-P project",
    author_email="extra-p@lists.parallel.informatik.tu-darmstadt.de",
    description="pyCubexR is a Python package for reading the Cube4 file format.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/extra-p/pycubexr",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires='>=3.7',
    install_requires=[
        "numpy~=1.18"
    ]
)



import unittest
from pathlib import Path

from pycubexr import CubexParser
from pycubexr.utils.exceptions import MissingMetricError


class TestMetricValuesKripke(unittest.TestCase):
    cubex: CubexParser = None

    @classmethod
    def setUpClass(cls) -> None:
        cubex_file_path = Path("../data/kripke.p8.d2.g32.r1/profile.cubex").resolve()
        cls.cubex = CubexParser(cubex_file_path).__enter__()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cubex.__exit__(None, None, None)

    def get_values_for_metric(self, name, convert_to_exclusive=False, convert_to_inclusive=False):
        metric = self.cubex.get_metric_by_name(name)
        metric_values = self.cubex.get_metric_values(metric)
        # iterate over all callpaths in cubex file
        cnode_values = []
        for cnode in self.cubex.all_cnodes():
            # return the measurement values for all mpi processes for the current metric and callpath
            cnode_values.append(metric_values.value(cnode, convert_to_exclusive, convert_to_inclusive))
        return cnode_values

    def test_visits_exclusive(self):
        correct_values = [8, 8, 33, 16, 8, 8000, 8000, 8000, 8000, 96000, 169025, 96000, 8000, 8, ]
        metric_values = self.get_values_for_metric('visits', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_visits_inclusive(self):
        correct_values = [401106, 8, 33, 16, 401033, 8000, 8000, 385025, 8000, 96000, 169025, 96000, 8000, 8, ]
        metric_values = self.get_values_for_metric('visits', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_time_exclusive(self):
        correct_values = [0.049351889999996956, 0.41779200749999995, 4.7357375E-4, 1.7047999999999998E-4,
                          0.4765991312500262, 59.994445453124996, 59.901360724375, 18.81082184375, 0.0791653,
                          1.9761004281249999, 3.677575010625, 2.70347292125, 0.5036856562500001, 0.04049549125, ]
        metric_values = self.get_values_for_metric('time', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_time_inclusive(self):
        correct_values = [148.63150991125, 0.41779200749999995, 4.7357375E-4, 1.7047999999999998E-4, 148.12322646875,
                          59.994445453124996, 59.901360724375, 27.750821159999997, 0.0791653, 1.9761004281249999,
                          3.677575010625, 2.70347292125, 0.5036856562500001, 0.04049549125, ]
        metric_values = self.get_values_for_metric('time', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_min_time_exclusive(self):
        correct_values = [18.57139089625, 0.0446771925, 5.263125E-6, 5.80125E-6, 18.5020277025, 0.00741907125,
                          0.00703084, 0.00325334125, 7.605E-6, 1.7570625E-5, 7.595E-6, 2.031375E-5, 5.414375E-5,
                          4.776875E-5, ]
        metric_values = self.get_values_for_metric('min_time', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_min_time_inclusive(self):
        correct_values = [5.263125E-6, 0.0446771925, 5.263125E-6, 5.80125E-6, 7.595E-6, 0.00741907125, 0.00703084,
                          7.595E-6, 7.605E-6, 1.7570625E-5, 7.595E-6, 2.031375E-5, 5.414375E-5, 4.776875E-5, ]
        metric_values = self.get_values_for_metric('min_time', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_max_time_exclusive(self):
        correct_values = [18.60063626375, 0.073913345, 6.36525E-5, 3.40825E-5, 18.51830332, 0.008342615, 0.00768267875,
                          0.0177915775, 1.236375E-5, 6.542625E-5, 1.2951E-4, 7.841E-5, 8.488E-5, 0.00588251125, ]
        metric_values = self.get_values_for_metric('max_time', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_max_time_inclusive(self):
        correct_values = [18.60063626375, 0.073913345, 6.36525E-5, 3.40825E-5, 18.51830332, 0.008342615, 0.00768267875,
                          0.0177915775, 1.236375E-5, 6.542625E-5, 1.2951E-4, 7.841E-5, 8.488E-5, 0.00588251125, ]
        metric_values = self.get_values_for_metric('max_time', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_task_migration_loss(self):
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'task_migration_loss',
                          convert_to_exclusive=True)
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'task_migration_loss',
                          convert_to_inclusive=True)

    def test_task_migration_win(self):
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'task_migration_win',
                          convert_to_exclusive=True)
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'task_migration_win',
                          convert_to_inclusive=True)

    def test_bytes_put(self):
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'bytes_put', convert_to_exclusive=True)
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'bytes_put', convert_to_inclusive=True)

    def test_bytes_get(self):
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'bytes_get', convert_to_exclusive=True)
        self.assertRaises(MissingMetricError, self.get_values_for_metric, 'bytes_get', convert_to_inclusive=True)

    def test_PAPI_TOT_INS_exclusive(self):
        correct_values = [16380495, 218090062, 82702, 41913, 92374590, 18428764176, 18440603145, 4575020158, 20131404,
                          550715334, 919093540, 580021611, 131607814, 9039208, ]
        metric_values = self.get_values_for_metric('PAPI_TOT_INS', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PAPI_TOT_INS_inclusive(self):
        correct_values = [43981966152, 218090062, 82702, 41913, 43738331772, 18428764176, 18440603145, 6776589861,
                          20131404, 550715334, 919093540, 580021611, 131607814, 9039208, ]
        metric_values = self.get_values_for_metric('PAPI_TOT_INS', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PAPI_FP_INS_exclusive(self):
        correct_values = [2135, 496, 0, 0, 144000, 2764800000, 2764800000, 1109152000, 0, 0, 0, 0, 0, 0, ]
        metric_values = self.get_values_for_metric('PAPI_FP_INS', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PAPI_FP_INS_inclusive(self):
        correct_values = [6638898631, 496, 0, 0, 6638896000, 2764800000, 2764800000, 1109152000, 0, 0, 0, 0, 0, 0, ]
        metric_values = self.get_values_for_metric('PAPI_FP_INS', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PAPI_FP_OPS_exclusive(self):
        correct_values = [2183, 480, 0, 0, 144000, 5529600000, 5529600000, 2697987161, 0, 0, 0, 0, 0, 0, ]
        metric_values = self.get_values_for_metric('PAPI_FP_OPS', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PAPI_FP_OPS_inclusive(self):
        correct_values = [13757333824, 480, 0, 0, 13757331161, 5529600000, 5529600000, 2697987161, 0, 0, 0, 0, 0, 0, ]
        metric_values = self.get_values_for_metric('PAPI_FP_OPS', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PEVT_L2_FETCH_LINE_exclusive(self):
        correct_values = [236294, 301627, 2475, 466, 9951320, 63665366, 621100937, 72281973, 484451, 14599752, 16286754,
                          7812501, 1204831, 8663, ]
        metric_values = self.get_values_for_metric('PEVT_L2_FETCH_LINE', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PEVT_L2_FETCH_LINE_inclusive(self):
        correct_values = [807937410, 301627, 2475, 466, 807387885, 63665366, 621100937, 112670262, 484451, 14599752,
                          16286754, 7812501, 1204831, 8663, ]
        metric_values = self.get_values_for_metric('PEVT_L2_FETCH_LINE', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PEVT_L2_STORE_LINE_exclusive(self):
        correct_values = [2279935, 2075179, 6347, 509, 12419687, 131884217, 761931928, 456598356, 980669, 47246817,
                          101259237, 55485219, 22286633, 6776, ]
        metric_values = self.get_values_for_metric('PEVT_L2_STORE_LINE', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_PEVT_L2_STORE_LINE_inclusive(self):
        correct_values = [1594461509, 2075179, 6347, 509, 1590092763, 131884217, 761931928, 683856931, 980669, 47246817,
                          101259237, 55485219, 22286633, 6776, ]
        metric_values = self.get_values_for_metric('PEVT_L2_STORE_LINE', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_bytes_sent_exclusive(self):
        correct_values = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1770240000, 0, 0, ]
        metric_values = self.get_values_for_metric('bytes_sent', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_bytes_sent_inclusive(self):
        correct_values = [1770240000, 0, 0, 0, 1770240000, 0, 0, 1770240000, 0, 0, 0, 1770240000, 0, 0, ]
        metric_values = self.get_values_for_metric('bytes_sent', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_bytes_received_exclusive(self):
        correct_values = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1770240000, 0, 0, 0, ]
        metric_values = self.get_values_for_metric('bytes_received', convert_to_exclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)

    def test_bytes_received_inclusive(self):
        correct_values = [1770240000, 0, 0, 0, 1770240000, 0, 0, 1770240000, 0, 0, 1770240000, 0, 0, 0, ]
        metric_values = self.get_values_for_metric('bytes_received', convert_to_inclusive=True)
        for correct, cnode_values in zip(correct_values, metric_values):
            self.assertAlmostEqual(correct, cnode_values)


if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path

from pycubexr import CubexParser
from pycubexr.classes.metric import MetricType
from pycubexr.utils.exceptions import MissingMetricError


class AbstractTestcase:
    class Mapping(unittest.TestCase):
        expected_mapping_visits: list
        expected_mapping_time: list
        path: str = None

        @classmethod
        def setUpClass(cls) -> None:
            if cls.path is None:
                raise unittest.SkipTest
            cubex_file_path = Path(cls.path).resolve()
            cls.cubex = CubexParser(cubex_file_path).__enter__()

        @classmethod
        def tearDownClass(cls) -> None:
            cls.cubex.__exit__(None, None, None)

        def test_visits(self):
            metric = self.cubex.get_metric_by_name('visits')
            self.assertEqual(MetricType.EXCLUSIVE, metric.metric_type)
            self.perform_test(metric, self.expected_mapping_visits)

        def test_time(self):
            metric = self.cubex.get_metric_by_name('time')
            self.assertEqual(MetricType.INCLUSIVE, metric.metric_type)
            self.perform_test(metric, self.expected_mapping_time)

        def test_other_metrics(self):
            for metric in self.cubex.get_metrics():
                if metric.metric_type == MetricType.EXCLUSIVE:
                    expected_mapping = self.expected_mapping_visits
                elif metric.metric_type == MetricType.INCLUSIVE:
                    expected_mapping = self.expected_mapping_time
                else:
                    self.fail("Unknown metric type")

                tree_index_to_cid_map = metric.tree_index_to_cid_map
                try:
                    cnode_indices = self.cubex.get_metric_values(metric, False).cnode_indices
                except MissingMetricError:
                    return
                for index, cid, cname in expected_mapping:
                    self.assertEqual(cid, tree_index_to_cid_map[index])
                    if cid in cnode_indices:
                        self.assertEqual(index, cnode_indices[cid], msg=str((metric.name, index, cid, cname)))
                    self.assertEqual(cname, self.cubex.get_cnode(cid).region.mangled_name)

        def perform_test(self, metric, expected_mapping):
            tree_index_to_cid_map = metric.tree_index_to_cid_map
            try:
                cnode_indices = self.cubex.get_metric_values(metric, False).cnode_indices
            except MissingMetricError:
                return
            self.assertEqual(len(cnode_indices), len(expected_mapping))
            for index, cid, cname in expected_mapping:
                self.assertEqual(cid, tree_index_to_cid_map[index])
                # the following is only true for metrics that have data for all call paths
                self.assertEqual(index, cnode_indices[cid], msg=str((metric.name, index, cid, cname)))
                self.assertEqual(cname, self.cubex.get_cnode(cid).region.mangled_name)


class TestSimpleCallTree(AbstractTestcase.Mapping):
    path = "../data/call_tree_test/call_tree_test.cubex"
    expected_mapping_visits = [
        # index,cnode_id,cnode_mangled_name
        (0, 0, "cube_test_file/test.x"),
        (1, 1, "main"),
        (2, 2, "a"),
        (3, 3, "a1"),
        (4, 4, "a2"),
        (5, 5, "a3"),
        (6, 6, "b"),
        (7, 7, "b1"),
        (8, 8, "b2"),
        (9, 9, "b3"),
        (10, 10, "c"),
        (11, 11, "c1"),
        (12, 12, "c2"),
        (13, 13, "c3"),
        (14, 14, "d"),
        (15, 15, "d1"),
        (16, 16, "d2"),
        (17, 17, "d3"),
    ]
    expected_mapping_time = [
        # index,cnode_id,cnode_mangled_name
        (0, 0, "cube_test_file/test.x"),
        (1, 1, "main"),
        (2, 2, "a"),
        (3, 6, "b"),
        (4, 10, "c"),
        (5, 14, "d"),
        (6, 3, "a1"),
        (7, 4, "a2"),
        (8, 5, "a3"),
        (9, 7, "b1"),
        (10, 8, "b2"),
        (11, 9, "b3"),
        (12, 11, "c1"),
        (13, 12, "c2"),
        (14, 13, "c3"),
        (15, 15, "d1"),
        (16, 16, "d2"),
        (17, 17, "d3"),
    ]


class TestProblematicCallTree(AbstractTestcase.Mapping):
    path = "../data/time.p4.n2000.x1.r0/profile.cubex"
    expected_mapping_visits = [
        # index,cnode_id,cnode_mangled_name
        (0, 0, "/work/home/ga42biko/new_codes/Benchmark_V2/A8/Test_A/bg_time -n 2000 -x 1"),
        (1, 1, "main"),
        (2, 2, "MPI_Init"),
        (3, 3, "MPI_Comm_rank"),
        (4, 4, "MPI_Comm_size"),
        (5, 5, "_Z7MODEL_1mmii"),
        (6, 6, "_ZN2bg6VectorIdEC2Em"),
        (7, 7, "_ZN2bg10DataStructIdEC2Ev"),
        (8, 8, "_ZN2bg6VectorIdEclEi"),
        (9, 9, "_ZN2bg8function3F_0IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (10, 10, "MPI_Bcast"),
        (11, 11, "_ZN2bg6VectorIdEclEi"),
        (12, 12, "_ZN2bg8function3F_1IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (13, 13, "MPI_Scatter"),
        (14, 14, "_ZN2bg6VectorIdEclEi"),
        (15, 15, "_ZN2bg8function3F_2IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (16, 16, "MPI_Allgather"),
        (17, 17, "_ZN2bg6VectorIdEclEi"),
        (18, 18, "_ZN2bg8function3F_3IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (19, 19, "MPI_Reduce"),
        (20, 20, "_ZN2bg6VectorIdEclEi"),
        (21, 21, "_ZN2bg8function3F_4IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (22, 22, "MPI_Allreduce"),
        (23, 23, "_ZN2bg6VectorIdEclEi"),
        (24, 24, "_ZN2bg8function3F_5IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (25, 25, "MPI_Bcast"),
        (26, 26, "_ZN2bg6VectorIdEclEi"),
        (27, 27, "_ZN2bg8function3F_6IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (28, 28, "MPI_Scatter"),
        (29, 29, "_ZN2bg6VectorIdEclEi"),
        (30, 30, "_ZN2bg8function3F_7IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (31, 31, "MPI_Gather"),
        (32, 32, "_ZN2bg6VectorIdEclEi"),
        (33, 33, "_ZN2bg8function3F_8IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (34, 34, "MPI_Send"),
        (35, 35, "MPI_Barrier"),
        (36, 36, "_ZN2bg6VectorIdEclEi"),
        (37, 44, "MPI_Recv"),
        (38, 37, "_ZN2bg8function3F_9IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (39, 38, "MPI_Send"),
        (40, 39, "MPI_Barrier"),
        (41, 40, "_ZN2bg6VectorIdEclEi"),
        (42, 45, "MPI_Recv"),
        (43, 41, "_ZN2bg6VectorIdED2Ev"),
        (44, 42, "_ZN2bg10DataStructIdED2Ev"),
        (45, 43, "MPI_Finalize"),
    ]
    expected_mapping_time = [
        # index,cnode_id,cnode_mangled_name
        (0, 0, "/work/home/ga42biko/new_codes/Benchmark_V2/A8/Test_A/bg_time -n 2000 -x 1"),
        (1, 1, "main"),
        (2, 2, "MPI_Init"),
        (3, 3, "MPI_Comm_rank"),
        (4, 4, "MPI_Comm_size"),
        (5, 5, "_Z7MODEL_1mmii"),
        (6, 43, "MPI_Finalize"),
        (7, 6, "_ZN2bg6VectorIdEC2Em"),
        (8, 8, "_ZN2bg6VectorIdEclEi"),
        (9, 9, "_ZN2bg8function3F_0IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (10, 12, "_ZN2bg8function3F_1IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (11, 15, "_ZN2bg8function3F_2IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (12, 18, "_ZN2bg8function3F_3IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (13, 21, "_ZN2bg8function3F_4IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (14, 24, "_ZN2bg8function3F_5IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (15, 27, "_ZN2bg8function3F_6IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (16, 30, "_ZN2bg8function3F_7IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (17, 33, "_ZN2bg8function3F_8IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (18, 37, "_ZN2bg8function3F_9IdEEvmmRNS_6VectorIT_EES5_S5_S5_mmii"),
        (19, 41, "_ZN2bg6VectorIdED2Ev"),
        (20, 7, "_ZN2bg10DataStructIdEC2Ev"),
        (21, 10, "MPI_Bcast"),
        (22, 11, "_ZN2bg6VectorIdEclEi"),
        (23, 13, "MPI_Scatter"),
        (24, 14, "_ZN2bg6VectorIdEclEi"),
        (25, 16, "MPI_Allgather"),
        (26, 17, "_ZN2bg6VectorIdEclEi"),
        (27, 19, "MPI_Reduce"),
        (28, 20, "_ZN2bg6VectorIdEclEi"),
        (29, 22, "MPI_Allreduce"),
        (30, 23, "_ZN2bg6VectorIdEclEi"),
        (31, 25, "MPI_Bcast"),
        (32, 26, "_ZN2bg6VectorIdEclEi"),
        (33, 28, "MPI_Scatter"),
        (34, 29, "_ZN2bg6VectorIdEclEi"),
        (35, 31, "MPI_Gather"),
        (36, 32, "_ZN2bg6VectorIdEclEi"),
        (37, 34, "MPI_Send"),
        (38, 35, "MPI_Barrier"),
        (39, 36, "_ZN2bg6VectorIdEclEi"),
        (40, 44, "MPI_Recv"),
        (41, 38, "MPI_Send"),
        (42, 39, "MPI_Barrier"),
        (43, 40, "_ZN2bg6VectorIdEclEi"),
        (44, 45, "MPI_Recv"),
        (45, 42, "_ZN2bg10DataStructIdED2Ev"),
    ]
    expected_bytes_sent = [
        # tree_index,cnode_id,cnode_mangled_name,file_index
        (10, 10, "MPI_Bcast", 0),
        (13, 13, "MPI_Scatter", 1),
        (16, 16, "MPI_Allgather", 2),
        (19, 19, "MPI_Reduce", 3),
        (22, 22, "MPI_Allreduce", 4),
        (25, 25, "MPI_Bcast", 5),
        (28, 28, "MPI_Scatter", 6),
        (31, 31, "MPI_Gather", 7),
        (34, 34, "MPI_Send", 8),
        (39, 38, "MPI_Send", 9),
    ]

    expected_bytes_received = [
        # tree_index,cnode_id,cnode_mangled_name,file_index
        (10, 10, "MPI_Bcast", 0),
        (13, 13, "MPI_Scatter", 1),
        (16, 16, "MPI_Allgather", 2),
        (19, 19, "MPI_Reduce", 3),
        (22, 22, "MPI_Allreduce", 4),
        (25, 25, "MPI_Bcast", 5),
        (28, 28, "MPI_Scatter", 6),
        (31, 31, "MPI_Gather", 7),
        (37, 44, "MPI_Recv", 8),
        (42, 45, "MPI_Recv", 9),
    ]

    def test_bytes_sent(self):
        metric = self.cubex.get_metric_by_name('bytes_sent')
        self.assertEqual(MetricType.EXCLUSIVE, metric.metric_type)
        tree_index_to_cid_map = metric.tree_index_to_cid_map
        cnode_indices = self.cubex.get_metric_values(metric, False).cnode_indices
        self.assertEqual(len(cnode_indices), len(self.expected_bytes_sent))
        for index, cid, cname, findex in self.expected_bytes_sent:
            self.assertEqual(cid, tree_index_to_cid_map[index])
            self.assertEqual(findex, cnode_indices[cid], msg=str((metric.name, index, cid, cname)))
            self.assertEqual(cname, self.cubex.get_cnode(cid).region.mangled_name)

    def test_bytes_recieved(self):
        metric = self.cubex.get_metric_by_name('bytes_received')
        self.assertEqual(MetricType.EXCLUSIVE, metric.metric_type)
        tree_index_to_cid_map = metric.tree_index_to_cid_map
        cnode_indices = self.cubex.get_metric_values(metric, False).cnode_indices
        self.assertEqual(len(cnode_indices), len(self.expected_bytes_received))
        for index, cid, cname, findex in self.expected_bytes_received:
            self.assertEqual(cid, tree_index_to_cid_map[index])
            self.assertEqual(findex, cnode_indices[cid], msg=str((metric.name, index, cid, cname)))
            self.assertEqual(cname, self.cubex.get_cnode(cid).region.mangled_name)


class TestBlast(AbstractTestcase.Mapping):
    path = "../data/blast.p64.r1/profile.cubex"

    expected_mapping_visits = [
        # index,cnode_id,cnode_mangled_name
        (0, 0, "PARALLEL"),
        (1, 1, "MPI_Init"),
        (2, 2, "MPI_Comm_size"),
        (3, 3, "MPI_Comm_rank"),
        (4, 4, "MPI_Bcast"),
        (5, 5, "MPI_Isend"),
        (6, 6, "MPI_Waitall"),
        (7, 7, "MPI_Allgather"),
        (8, 8, "MPI_Reduce"),
        (9, 9, "MPI_Allreduce"),
        (10, 10, "MPI_Irecv"),
        (11, 11, "MPI_Allgatherv"),
        (12, 12, "ComputeCornerForces"),
        (13, 13, "MPI_Reduce"),
        (14, 14, "main"),
        (15, 15, "MPI_Comm_rank"),
        (16, 16, "MPI_Comm_size"),
        (17, 17, "MPI_Irecv"),
        (18, 18, "MPI_Waitall"),
        (19, 19, "MPI_Isend"),
        (20, 20, "Eval_dv_dt"),
        (21, 21, "MPI_Comm_rank"),
        (22, 22, "MPI_Comm_size"),
        (23, 23, "MPI_Irecv"),
        (24, 24, "MPI_Isend"),
        (25, 25, "MPI_Waitall"),
        (26, 26, "MPI_Allreduce"),
        (27, 27, "ComputeCornerForces"),
        (28, 28, "MPI_Reduce"),
        (29, 29, "MPI_Allreduce"),
        (30, 30, "MPI_Finalize"),
        (31, 31, "MPI_Recv"),
    ]
    expected_mapping_time = [
        # index,cnode_id,cnode_mangled_name
        (0, 0, "PARALLEL"),
        (1, 1, "MPI_Init"),
        (2, 2, "MPI_Comm_size"),
        (3, 3, "MPI_Comm_rank"),
        (4, 4, "MPI_Bcast"),
        (5, 5, "MPI_Isend"),
        (6, 6, "MPI_Waitall"),
        (7, 7, "MPI_Allgather"),
        (8, 8, "MPI_Reduce"),
        (9, 9, "MPI_Allreduce"),
        (10, 10, "MPI_Irecv"),
        (11, 11, "MPI_Allgatherv"),
        (12, 12, "ComputeCornerForces"),
        (13, 14, "main"),
        (14, 30, "MPI_Finalize"),
        (15, 31, "MPI_Recv"),
        (16, 13, "MPI_Reduce"),
        (17, 15, "MPI_Comm_rank"),
        (18, 16, "MPI_Comm_size"),
        (19, 17, "MPI_Irecv"),
        (20, 18, "MPI_Waitall"),
        (21, 19, "MPI_Isend"),
        (22, 20, "Eval_dv_dt"),
        (23, 27, "ComputeCornerForces"),
        (24, 29, "MPI_Allreduce"),
        (25, 21, "MPI_Comm_rank"),
        (26, 22, "MPI_Comm_size"),
        (27, 23, "MPI_Irecv"),
        (28, 24, "MPI_Isend"),
        (29, 25, "MPI_Waitall"),
        (30, 26, "MPI_Allreduce"),
        (31, 28, "MPI_Reduce"),
    ]
